import asyncio

from asyncio.futures import CancelledError
from aioetcd3.base import StubMixin
from aioetcd3._etcdv3 import rpc_pb2 as rpc
from aioetcd3._etcdv3 import kv_pb2 as kv
from asyncio.queues import Queue, QueueEmpty, QueueFull
from aioetcd3.utils import put_key_range
from aioetcd3.kv import KVMetadata
import aioetcd3._etcdv3.rpc_pb2_grpc as stub

EVENT_TYPE_MODIFY = "MODIFY"
EVENT_TYPE_DELETE = "DELETE"
EVENT_TYPE_CREATE = "CREATE"


class Event(object):
    def __init__(self, event, revision):
        if event.type == kv.Event.PUT:
            if event.kv.version == 1:
                self.type = EVENT_TYPE_CREATE
            else:
                self.type = EVENT_TYPE_MODIFY
        else:
            self.type = EVENT_TYPE_DELETE

        self.key = event.kv.key
        self.value = event.kv.value
        self.meta = KVMetadata(event.kv)

        self.pre_value = event.prev_kv.value
        self.pre_meta = KVMetadata(event.prev_kv)
        
        self.revision = revision

    def is_put(self):
        return self.type == EVENT_TYPE_CREATE or self.type == EVENT_TYPE_MODIFY

    def __str__(self):
        return f'{self.type} {self.key},{self.value}'

        
class WatchScope(object):
    def __init__(self, _iter):
        self._iter = _iter

    async def __aenter__(self):
        await self._iter.__anext__()
        return self._iter

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        try:
            await self._iter.aclose()
        except GeneratorExit:
            pass


class _Pipe(object):
    """
    Selectable asyncio channel
    """
    def __init__(self, maxsize=None, *, loop=None):
        self._loop = loop
        self._notify = asyncio.Event(loop=loop)
        self._full_notify = asyncio.Event(loop=loop)
        self._queue = []
        self._full_notify.set()
        if maxsize is None or maxsize <= 0:
            self._maxsize = None
        else:
            self._maxsize = maxsize
        self._last_watch_version = None
    
    def is_empty(self):
        return not self._notify.is_set()
    
    def is_full(self):
        return not self._full_notify.is_set()
    
    async def put(self, value):
        await self.wait_full()
        self.put_nowait(value)
    
    def put_nowait(self, value):
        if self.is_full():
            raise QueueFull
        self._queue.append(value)
        if self.is_empty():
            self._notify.set()
        if self._maxsize and len(self._queue) >= self._maxsize:
            self._full_notify.clear()        
    
    async def get(self, value):
        await self.wait_empty()
        return self.get_nowait()
    
    def get_nowait(self):
        if self.is_empty():
            raise QueueEmpty
        if self._maxsize or len(self._queue) <= self._maxsize:
            self._full_notify.set()
        if len(self._queue) == 1:
            self._notify.clear()
        return self._queue.pop(0)
    
    async def read(self, limit=None):
        await self.wait_empty()
        return self.read_nowait(limit)
    
    def read_nowait(self, limit=None):
        if self.is_empty():
            raise QueueEmpty
        if limit is None or limit <= 0:
            read_size = len(self._queue)
        else:
            read_size = min(len(self._queue), limit)
        result = self._queue[:read_size]
        del self._queue[:read_size]
        if not self._maxsize or len(self._queue) < self._maxsize:
            self._full_notify.set()
        if len(self._queue) == 0:
            self._notify.clear()
        return result
    
    async def write(self, values):
        await self.wait_full()
        return self.write_nowait(values)
    
    def write_nowait(self, values):
        if self.is_full():
            raise QueueFull
        if self._maxsize is None:
            write_size = len(values)
        else:
            write_size = min(len(values), self._maxsize - len(self._queue))
        self._queue.extend(values[:write_size])
        if len(self._queue) > 0:
            self._notify.set()
        if self._maxsize and len(self._queue) >= self._maxsize:
            self._full_notify.clear()
        return write_size
    
    async def wait_full(self):
        while self.is_full():
            await self._full_notify.wait()
        
    async def wait_empty(self):
        while self.is_empty():
            await self._notify.wait()


async def _select(pipes, futures, *, loop=None):
    futures = [asyncio.ensure_future(f, loop=loop) for f in futures]
    _, pending = await asyncio.wait([p.wait_empty() for p in pipes] + list(futures),
                        loop=loop, return_when=asyncio.FIRST_COMPLETED)
    for p in pending:
        if p not in futures:
            p.cancel()
            try:
                await p
            except Exception:
                pass
    return [p for p in pipes if not p.is_empty()], [f for f in futures if f.done()]


class WatchException(Exception):
    def _clone(self):
        return type(self)(*self.args)

class CompactRevisonException(WatchException):
    def __init__(self, revision):
        super().__init__(f"Watch on compact revision. Min revision is {revision}")
        self.revision = revision
    
    def _clone(self):
        return CompactRevisonException(self.revision)


class ServerCancelException(WatchException):
    def __init__(self, cancel_reason):
        super().__init__(f"Watch cancelled: {cancel_reason}")
        self.cancel_reason = cancel_reason
    
    def _clone(self):
        return ServerCancelException(self.cancel_reason)


class Watch(StubMixin):
    def __init__(self, channel, timeout):
        # Put (WatchCreateRequest, output_queue, done_future) to create a watch
        self._create_request_queue = _Pipe(5, loop=self._loop)
        # Put (output_queue, done_future) to cancel a watch
        self._cancel_request_queue = _Pipe(loop=self._loop)
        self._reconnect_event = asyncio.Event(loop=self._loop)
        self._watch_task_running = None
        super().__init__(channel, timeout)
    
    async def _watch_task(self, reconnect_event):
        # Queue for WatchRequest
        async def input_iterator(input_queue):
            while True:
                n = await input_queue.get()
                if n is None:
                    break
                yield n
        async def watch_call(input_queue, watch_stub, output_pipe):
            async with watch_stub.Watch.with_scope(input_iterator(input_queue)) as response_iter:
                async for r in response_iter:
                    await output_pipe.put(r)
        last_received_revision = None
        # watch_id -> revision
        last_watches_revision = {}
        # watch_id -> (WatchCreateRequest, output_queue)
        registered_watches = {}
        # output_queue -> watch_id
        registered_queues = {}
        # A tuple (WatchCreateRequest, output_queue, done_future, cancel_future)
        pending_create_request = None
        # watch_id -> done_future
        pending_cancel_requests = {}
        # output_queue -> (WatchCreateRequest, done_future)
        restore_creates = {}
        quitting = False
        def _reconnect_revision(watch_id):
            if last_received_revision is None:
                return None
            else:
                if watch_id in last_watches_revision:
                    last_revision = last_watches_revision[watch_id]
                    return max(last_revision + 1, last_received_revision)
                else:
                    return None
        try:
            while not quitting:         # Auto reconnect when failed or channel updated
                reconnect_event.clear()
                output_pipe = _Pipe(loop=self._loop)
                input_queue = asyncio.Queue(loop=self._loop)
                call_task = asyncio.ensure_future(watch_call(input_queue, self._watch_stub, output_pipe), loop=self._loop)
                try:
                    # Restore registered watches
                    for watch_id, (create_request, output_queue) in registered_watches.items():
                        if watch_id in pending_cancel_requests:
                            # Already cancelling
                            fut = pending_cancel_requests.pop(watch_id)
                            if not fut.done():
                                fut.set_result(True)
                            continue
                        r = rpc.WatchCreateRequest()
                        r.CopyFrom(create_request)
                        restore_revision = _reconnect_revision(watch_id)
                        if restore_revision is not None:
                            r.start_revision = restore_revision
                        restore_creates[output_queue] = (r, None)
                    registered_watches.clear()
                    registered_queues.clear()
                    # Restore pending cancels - should already be processed though
                    for watch_id, fut in pending_cancel_requests.items():
                        fut.set_result(True)
                    pending_cancel_requests.clear()
                    # Restore pending create request
                    if pending_create_request is not None:
                        if pending_create_request[3] is not None:     # Cancelled
                            pending_create_request[1].put_nowait((False, None, None))
                            if pending_create_request[2] is not None and not pending_create_request[2].done():
                                pending_create_request[2].set_result(True)
                            if pending_create_request[3] is not None and not pending_create_request[3].done():
                                pending_create_request[3].set_result(True)
                        else:
                            restore_creates[pending_create_request[1]] = (pending_create_request[0], pending_create_request[2])
                        pending_create_request = None
                    while True:
                        if pending_create_request is None:
                            if restore_creates:
                                q, (req, fut) = restore_creates.popitem()
                                # Send create request
                                pending_create_request = (req, q, fut, None)
                                input_queue.put_nowait(rpc.WatchRequest(create_request=req))
                        if pending_create_request is None:
                            select_pipes = [output_pipe, self._create_request_queue, self._cancel_request_queue]
                        else:
                            select_pipes = [output_pipe, self._cancel_request_queue]
                        reconn_wait = asyncio.ensure_future(reconnect_event.wait(), loop=self._loop)
                        select_futs = [reconn_wait, call_task]
                        if not pending_create_request and not registered_watches and \
                                not restore_creates:
                            select_futs.append(asyncio.sleep(2, loop=self._loop))
                        pipes, _ = await _select(select_pipes,
                                                    select_futs,
                                                    loop=self._loop)
                        reconn_wait.cancel()
                        if not pipes and not reconnect_event.is_set() and not call_task.done():
                            # No watch, stop the task
                            quitting = True
                            break
                        # Process cancel requests first
                        if self._cancel_request_queue in pipes:
                            cancel_requests = self._cancel_request_queue.read_nowait()
                            for output_queue, done_fut in cancel_requests:
                                if output_queue in pending_cancel_requests:
                                    # Chain this future
                                    pending_cancel_requests[output_queue].add_done_callback(
                                        lambda f, done_fut=done_fut: done_fut.set_result(True)
                                    )
                                elif output_queue in restore_creates:
                                    # Cancel a request which is not started
                                    _, fut = restore_creates.pop(output_queue)
                                    output_queue.put_nowait((False, None, None))
                                    if fut is not None and not fut.done():
                                        fut.set_result(True)
                                    if done_fut is not None and not done_fut.done():
                                        done_fut.set_result(True)
                                elif pending_create_request is not None and \
                                        pending_create_request[1] == output_queue:
                                    # Cancel the pending create watch
                                    if pending_create_request[3] is None:
                                        pending_create_request = pending_create_request[:3] + (done_fut,)
                                    else:
                                        pending_create_request[3].add_done_callback(
                                            lambda f, done_fut=done_fut: done_fut.set_result(True))
                                else:
                                    watch_id = registered_queues.get(output_queue)
                                    if watch_id is None:
                                        done_fut.set_result(True)
                                    else:
                                        # Send cancel request and save it to pending requests
                                        input_queue.put_nowait(
                                                rpc.WatchRequest(
                                                    cancel_request=
                                                        rpc.WatchCancelRequest(watch_id=watch_id)
                                                    )
                                                )
                                        pending_cancel_requests[watch_id] = done_fut
                        # Process received events
                        if output_pipe in pipes:
                            outputs = output_pipe.read_nowait()
                            for response in outputs:
                                if response.created:
                                    assert pending_create_request is not None
                                    if response.compact_revision > 0:
                                        # Cancelled (Is it possible?)
                                        exc = CompactRevisonException(response.compact_revision)
                                        pending_create_request[1].put_nowait((False, exc, response.compact_revision))
                                        if pending_create_request[2] is not None and not \
                                                pending_create_request[2].done():
                                            pending_create_request[2].set_exception(exc)
                                        if pending_create_request[3] is not None and not \
                                                pending_create_request[3].done():
                                            pending_create_request[3].set_result(True)
                                    else:
                                        registered_watches[response.watch_id] = pending_create_request[0:2]
                                        registered_queues[pending_create_request[1]] = response.watch_id
                                        if pending_create_request[2] is not None and not \
                                                pending_create_request[2].done():
                                            pending_create_request[2].set_result(True)
                                        if pending_create_request[3] is not None:
                                            # Immediately cancel the watch
                                            input_queue.put_nowait(
                                                    rpc.WatchRequest(
                                                        cancel_request=
                                                            rpc.WatchCancelRequest(watch_id=response.watch_id)
                                                        )
                                                    )
                                            pending_cancel_requests[response.watch_id] = pending_create_request[3]
                                    pending_create_request = None
                                if response.events:
                                    last_received_revision = response.header.revision
                                    last_watches_revision[response.watch_id] = last_received_revision
                                    if response.watch_id in registered_watches:
                                        _, output_queue = registered_watches[response.watch_id]
                                        output_queue.put_nowait((True,
                                                                 [Event(e, last_received_revision) for e in response.events],
                                                                 last_received_revision))
                                if response.compact_revision > 0:
                                    if response.watch_id in registered_watches:
                                        _, output_queue = registered_watches.pop(response.watch_id)
                                        exc = CompactRevisonException(response.compact_revision)
                                        output_queue.put_nowait((False, exc, response.compact_revision))
                                        del registered_queues[output_queue]
                                    if response.watch_id in pending_cancel_requests:
                                        if not pending_cancel_requests[response.watch_id].done():
                                            pending_cancel_requests[response.watch_id].set_result(True)
                                        del pending_cancel_requests[response.watch_id]
                                if response.canceled:
                                    # Cancel response
                                    if response.watch_id in registered_watches:
                                        _, output_queue = registered_watches.pop(response.watch_id)
                                        if response.watch_id in pending_cancel_requests:
                                            # Normal cancel
                                            output_queue.put_nowait((False, None, None))
                                        else:
                                            output_queue.put_nowait((False, ServerCancelException(response.cancel_reason), _reconnect_revision(response.watch_id)))
                                        del registered_queues[output_queue]
                                    if response.watch_id in pending_cancel_requests:
                                        if not pending_cancel_requests[response.watch_id].done():
                                            pending_cancel_requests[response.watch_id].set_result(True)
                                        del pending_cancel_requests[response.watch_id]
                        if self._create_request_queue in pipes:
                            while pending_create_request is None and not self._create_request_queue.is_empty():
                                create_req, output_queue, done_fut = self._create_request_queue.get_nowait()
                                if done_fut.done():
                                    # Ignore cancelled create requests
                                    output_queue.put_nowait((False, None, None))
                                    continue
                                # Send create request
                                pending_create_request = (create_req, output_queue, done_fut, None)
                                input_queue.put_nowait(rpc.WatchRequest(create_request=create_req))
                        if reconnect_event.is_set():
                            # Reconnected
                            break
                        if call_task.done():
                            # Maybe not available
                            if call_task.exception() is not None:
                                await call_task
                            else:
                                break
                finally:
                    input_queue.put_nowait(None)
                    call_task.cancel()
                    if quitting:
                        self._watch_task_running = None
                    try:
                        await call_task
                    except Exception:
                        pass
        except Exception as exc:
            if registered_queues:
                for q, watch_id in registered_queues.items():
                    
                    q.put_nowait((False, exc, _reconnect_revision(watch_id)))
            if pending_create_request is not None:
                pending_create_request[1].put_nowait((False, exc, None))
                if pending_create_request[2] is not None and not pending_create_request[2].done():
                    pending_create_request[2].set_exception(exc)
                if pending_create_request[3] is not None and not pending_create_request[3].done():
                    pending_create_request[3].set_result(True)
            if pending_cancel_requests:
                for _, fut in pending_cancel_requests.items():
                    if not fut.done():
                        fut.set_result(True)
            if restore_creates:
                for q, (_, fut) in restore_creates.items():
                    if fut is not None and not fut.done():
                        fut.set_result(exc)
                    q.put_nowait((False, exc, _reconnect_revision(watch_id)))
            if not self._create_request_queue.is_empty():
                create_requests = self._create_request_queue.read_nowait()
                for r in create_requests:
                    r[1].put_nowait((False, exc, None))
                    if r[2] is not None and not r[2].done():
                        r[2].set_exception(exc)
            if not self._cancel_request_queue.is_empty():
                cancel_requests = self._cancel_request_queue.read_nowait()
                for _, fut in cancel_requests:
                    if fut is not None and not fut.done():
                        fut.set_result(True)

            if exc is CancelledError:
                raise

        finally:
            self._watch_task_running = None
    
    def _update_channel(self, channel):
        super()._update_channel(channel)
        self._watch_stub = stub.WatchStub(channel)
        self._reconnect_event.set()

    def _ensure_watch_task(self):
        if self._watch_task_running is None:
            self._watch_task_running = asyncio.ensure_future(self._watch_task(self._reconnect_event))
    
    async def watch(self, key_range, start_revision=None, noput=False, nodelete=False, prev_kv=False,
                    always_reconnect=False, ignore_compact=False, batch_events=False, create_event=False):

        filters = []
        if noput:
            filters.append(rpc.WatchCreateRequest.NOPUT)
        if nodelete:
            filters.append(rpc.WatchCreateRequest.NODELETE)
        reconnect_revision = start_revision
        done_future = None
        try:
            while True:
                watch_request = rpc.WatchCreateRequest(start_revision=reconnect_revision,
                                                       filters=filters,
                                                       prev_kv=prev_kv)
        
                put_key_range(watch_request, key_range)
                self._ensure_watch_task()
                output_queue = asyncio.Queue(loop=self._loop)
                done_future = self._loop.create_future()
                await self._create_request_queue.put((watch_request, output_queue, done_future))
                try:
                    await done_future
                    if create_event:
                        yield None
                        create_event = False
                    while True:
                        is_event, result, revision = await output_queue.get()
                        if not is_event:
                            if revision is not None:
                                reconnect_revision = revision
                            if result is None:
                                break
                            else:
                                # When an exception is raised in multiple positions
                                # the traceback will mix up, so clone the exception
                                # for each raise
                                if isinstance(result, WatchException):
                                    raise result._clone() from result
                                else:
                                    raise WatchException("Watch failed with server exception") from result
                        else:
                            reconnect_revision = revision + 1
                            if batch_events:
                                yield tuple(result)
                            else:
                                for e in result:
                                    yield e
                except CompactRevisonException:
                    if ignore_compact:
                        continue
                    else:
                        raise
                except CancelledError:
                    raise
                except Exception:
                    if always_reconnect:
                        continue
                    else:
                        raise
                else:
                    break
        finally:
            if done_future is not None and not done_future.done():
                done_future.cancel()
            if self._watch_task_running:
                done_future = self._loop.create_future()
                await self._cancel_request_queue.put((output_queue, done_future))
                if self._watch_task_running:
                    try:
                        await done_future
                    except Exception:
                        pass

    def watch_scope(self, key_range, start_revision=None, noput=False, nodelete=False, prev_kv=False,
                    always_reconnect=False, ignore_compact=False, batch_events=False):
        return WatchScope(self.watch(key_range, start_revision=start_revision,
                                     noput=noput, nodelete=nodelete, prev_kv=prev_kv, create_event=True,
                                     always_reconnect=always_reconnect, ignore_compact=ignore_compact,
                                     batch_events=batch_events))



from aioetcd3._etcdv3 import rpc_pb2 as rpc
from aioetcd3.utils import to_bytes


class BaseCompare(object):
    def __init__(self, key):
        self.key = key
        self.value = None
        self.op = None

    # Version, Mod and Create can only be ints
    def __eq__(self, other):
        self.value = other
        self.op = rpc.Compare.EQUAL
        return self

    def __ne__(self, other):
        self.value = other
        self.op = rpc.Compare.NOT_EQUAL
        return self

    def __lt__(self, other):
        self.value = other
        self.op = rpc.Compare.LESS
        return self

    def __gt__(self, other):
        self.value = other
        self.op = rpc.Compare.GREATER
        return self

    def __repr__(self):
        return "{}: {} {} '{}'".format(self.__class__, self.key, self.op, self.value)

    def build_message(self):
        compare = rpc.Compare()
        compare.key = to_bytes(self.key)

        if self.op is None:
            raise ValueError('op must be one of =, < or >')

        compare.result = self.op

        self.build_compare(compare)
        return compare
    
    def build_compare(self, compare):
        raise NotImplementedError


class Value(BaseCompare):
    def build_compare(self, compare):
        compare.target = rpc.Compare.VALUE
        compare.value = to_bytes(self.value)


class Version(BaseCompare):
    def build_compare(self, compare):
        compare.target = rpc.Compare.VERSION
        compare.version = int(self.value)


class Create(BaseCompare):
    def build_compare(self, compare):
        compare.target = rpc.Compare.CREATE
        compare.create_revision = int(self.value)


class Mod(BaseCompare):
    def build_compare(self, compare):
        compare.target = rpc.Compare.MOD
        compare.mod_revision = int(self.value)



_default_timeout = object()

class StubMixin(object):
    def __init__(self, channel, timeout):
        self.channel = channel
        self.timeout = timeout
        self.last_response_info = None
        self._update_channel(channel)

    def _update_channel(self, channel):
        self.channel = channel
        self._loop = channel._loop

    def _update_cluster_info(self, header):
        self.last_response_info = header

    def get_cluster_info(self):
        return self.last_response_info

    async def grpc_call(self, stub_func, request, timeout=_default_timeout):
        if timeout is _default_timeout:
            timeout = self.timeout
        response = await stub_func(request, timeout=timeout)
        self._update_cluster_info(response.header)
        return response



import unittest
import asyncio
import functools
from aioetcd3.client import client
from aioetcd3.kv import KV
from aioetcd3.help import range_all, range_prefix, range_greater, range_greater_equal
from aioetcd3 import transaction


def asynctest(f):
    @functools.wraps(f)
    def _f(self):
        return asyncio.get_event_loop().run_until_complete(f(self))

    return _f


class KVTest(unittest.TestCase):

    def setUp(self):
        endpoints = "127.0.0.1:2379"
        self.client = client(endpoint=endpoints)
        endpoints = "127.0.0.1:2379"
        self.client.update_server_list(endpoint=endpoints)
        self.tearDown()

    @asynctest
    async def tearDown(self):
        await self.client.delete(key_range=range_all())

    @asynctest
    async def test_put_get(self):

        for i in range(0, 10):
            key = '/test' + str(i)
            value, meta = await self.client.put(key, str(i))
            self.assertIsNone(value)
            self.assertIsNone(meta)

        value, meta = await self.client.put('/test9', "10", prev_kv=True)
        self.assertEqual(value, b'9')
        self.assertIsNotNone(meta)

        value, meta = await self.client.put('/test9', "9", prev_kv=True, ignore_value=True)
        self.assertEqual(value, b'10')
        self.assertIsNotNone(meta)

        value, meta = await self.client.put('/test9', "9", prev_kv=True)
        self.assertEqual(value, b'10')
        self.assertIsNotNone(meta)

        count = await self.client.count(key_range=range_all())
        self.assertEqual(count, 10)

        value, meta = await self.client.get("/test9")
        self.assertEqual(value, b'9')
        self.assertIsNotNone(meta)

        keys_list = await self.client.range_keys(key_range=range_all())
        self.assertEqual(len(keys_list), 10)

        value_list = await self.client.range(key_range=range_all())
        self.assertEqual(len(value_list), 10)
        value = [v[1].decode('utf-8') for v in value_list]
        value.sort()
        real_value = [str(i) for i in range(0, 10)]
        self.assertEqual(value, real_value)

        value_list = await self.client.range(key_range=range_all(), limit=5)
        self.assertEqual(len(value_list), 5)

        value_list = await self.client.range(key_range=range_prefix('/'))
        self.assertEqual(len(value_list), 10)

        value_list = await self.client.range(key_range=range_prefix('/'), limit=11)
        self.assertEqual(len(value_list), 10)

        value_list = await self.client.range(key_range=range_greater_equal('/test8'))
        self.assertEqual(len(value_list), 2)
        self.assertEqual(value_list[0][1], b'8')
        self.assertEqual(value_list[1][1], b'9')

        value_list = await self.client.range(key_range=range_greater('/testa'))
        self.assertEqual(len(value_list), 0)

        await self.client.delete(key_range='/test9')
        value, meta = await self.client.get("/test9")
        self.assertIsNone(value)
        self.assertIsNone(meta)

        value_list = await self.client.pop(key_range='/test8')
        self.assertEqual(len(value_list), 1)
        self.assertEqual(value_list[0][0], b'/test8')
        self.assertEqual(value_list[0][1], b'8')

        value_list = await self.client.delete(key_range=range_prefix('/'), prev_kv=True)
        self.assertEqual(len(value_list), 8)

    @asynctest
    async def test_transaction(self):
        await self.client.put('/trans1', 'trans1')
        await self.client.put('/trans2', 'trans2')
        
        is_success, response = await self.client.txn(compare=[
            transaction.Value('/trans1') == b'trans1',
            transaction.Value('/trans2') == b'trans2'
        ], success=[
            KV.get.txn('/trans1'),
            KV.range.txn('/trans2')
        ], fail=[
            KV.delete.txn('/trans1')
        ])

        self.assertEqual(is_success, True)
        self.assertEqual(len(response), 2)
        
        self.assertEqual(response[0][0], b'trans1')
        self.assertEqual(response[1][0][:2], (b'/trans2', b'trans2'))
        
        is_success, response = await self.client.txn(compare=[
            transaction.Value('/trans1') == b'trans1',
            transaction.Value('/trans2') == b'trans2'
        ], success=[
            KV.delete.txn('/trans1'),
            KV.put.txn('/trans2', 'trans2', prev_kv=True),
            KV.put.txn('/trans3', 'trans3', prev_kv=True)
        ], fail=[
            KV.delete.txn('/trans1')
        ])

        self.assertEqual(is_success, True)
        self.assertEqual(len(response), 3)
        del_response = response[0]
        self.assertEqual(del_response, 1)
        put_response = response[1]
        self.assertEqual(put_response[0], b'trans2')
        put_response = response[2]
        # there is not pre_kv None
        self.assertIsNone(put_response[0])

        is_success, response = await self.client.txn(compare=[
            transaction.Value('/trans3') != b'trans3',
            transaction.Version('/trans3') < 1000,
            transaction.Mod('/trans3') > 100,
            transaction.Create('/trans3') != 200
        ], success=[
        ], fail=[
            KV.delete.txn('/trans3', prev_kv=True)
        ])

        self.assertEqual(is_success, False)
        self.assertEqual(len(response), 1)
        self.assertEqual(len(response[0]), 1)
        self.assertEqual(response[0][0][:2], (b'/trans3', b'trans3'))

if __name__ == '__main__':
    unittest.main()




import functools
import aiogrpc
import grpc

from aioetcd3._etcdv3 import rpc_pb2 as rpc
from aioetcd3.base import StubMixin
import aioetcd3._etcdv3.rpc_pb2_grpc as stub
from aioetcd3.utils import ipv4_endpoints
from aioetcd3.maintenance import Maintenance


def call_grpc(request, response_func, method):

    def _f(f):
        @functools.wraps(f)
        async def call(self, *args, **kwargs):
            r = await self.grpc_call(method(self), request(*args, **kwargs))
            return response_func(r)

        return call

    return _f


class Cluster(StubMixin):
    def _update_channel(self, channel):
        super()._update_channel(channel)
        self._cluster_stub = stub.ClusterStub(channel)

    @call_grpc(lambda peerurls: rpc.MemberAddRequest(peerURLs=peerurls),
               lambda r: r.member, lambda s: s._cluster_stub.MemberAdd)
    async def member_add(self, peerurls):
        pass

    @call_grpc(lambda mid: rpc.MemberRemoveRequest(ID=mid),
               lambda r: [m for m in r.members],
               lambda s: s._cluster_stub.MemberRemove)
    async def member_remove(self, mid):
        pass

    @call_grpc(lambda mid, urls: rpc.MemberUpdateRequest(ID=mid, peerURLs=urls),
               lambda r: [m for m in r.members],
               lambda s: s._cluster_stub.MemberUpdate)
    async def member_update(self, mid, peerurls):
        pass

    @call_grpc(lambda: rpc.MemberListRequest(), lambda r: [m for m in r.members],
               lambda s: s._cluster_stub.MemberList)
    async def member_list(self):
        pass

    async def member_healthy(self, members=None):

        if not members:
            members = await self.member_list()
            members = [m.clientURLs for m in members]

        health_members = []
        unhealth_members = []
        for m in members:

            m = [u.rpartition("//")[2] for u in m]
            m = [u for u in m if u]
            if m:
                server_endpoint = ipv4_endpoints(m)
    
                if self._credentials:
                    channel = aiogrpc.secure_channel(server_endpoint, self._credentials, options=self._options,
                                                     loop=self._loop, executor=self._executor,
                                                     standalone_pool_for_streaming=True)
                else:
                    channel = aiogrpc.insecure_channel(server_endpoint, options=self._options, loop=self._loop,
                                                       executor=self._executor, standalone_pool_for_streaming=True)
                try:
                    maintenance = Maintenance(channel=channel, timeout=2)
                    try:
                        await maintenance.status()
                    except grpc.RpcError:
                        unhealth_members.append(m)
                    else:
                        health_members.append(m)
                finally:
                    await channel.close()
            else:
                unhealth_members.append(m)

        return health_members, unhealth_members





import unittest
import asyncio
import functools

from aioetcd3.client import client
from aioetcd3.help import range_all


def asynctest(f):
    @functools.wraps(f)
    def _f(self):
        asyncio.get_event_loop().run_until_complete(f(self))

    return _f


class LeaseTest(unittest.TestCase):
    def setUp(self):
        endpoints = "127.0.0.1:2379"
        self.client = client(endpoint=endpoints)

        self.tearDown()

    @asynctest
    async def test_lease_1(self):
        lease = await self.client.grant_lease(ttl=5)
        self.assertEqual(lease.ttl, 5)

        await asyncio.sleep(1)
        lease, keys = await self.client.get_lease_info(lease)
        self.assertLessEqual(lease.ttl, 4)
        self.assertEqual(len(keys), 0)

        lease = await self.client.refresh_lease(lease)
        self.assertEqual(lease.ttl, 5)

        await self.client.revoke_lease(lease)

        lease, keys = await self.client.get_lease_info(lease)
        self.assertIsNone(lease)
        self.assertEqual(len(keys), 0)

    @asynctest
    async def test_lease_2(self):
        lease = await self.client.grant_lease(ttl=5)
        self.assertEqual(lease.ttl, 5)

        await asyncio.sleep(1)
        lease, keys = await lease.info()
        self.assertLessEqual(lease.ttl, 4)
        self.assertEqual(len(keys), 0)

        lease = await lease.refresh()
        self.assertEqual(lease.ttl, 5)

        await lease.revoke()
        lease, keys = await lease.info()
        self.assertIsNone(lease)
        self.assertEqual(len(keys), 0)

        lease = None
        async with self.client.grant_lease_scope(ttl=5) as l:
            lease = l
            await asyncio.sleep(1)

        lease, keys = await lease.info()
        self.assertIsNone(lease)
        self.assertEqual(len(keys), 0)


    @asynctest
    async def test_lease_3(self):
        lease = await self.client.grant_lease(ttl=5)
        self.assertEqual(lease.ttl, 5)

        await self.client.put("/testlease", "testlease", lease=lease)

        await asyncio.sleep(6)
        lease, keys = await lease.info()
        self.assertIsNone(lease, None)
        self.assertEqual(len(keys), 0)

        value, meta = await self.client.get('/testlease')
        self.assertIsNone(value)
        self.assertIsNone(meta)

    @asynctest
    async def tearDown(self):
        await self.client.delete(range_all())



import functools

from aioetcd3.base import StubMixin
from aioetcd3._etcdv3 import rpc_pb2 as rpc
from aioetcd3._etcdv3 import auth_pb2 as auth
from aioetcd3.utils import put_key_range
import aioetcd3._etcdv3.rpc_pb2_grpc as stub


def call_grpc(request, response_func, method):

    def _f(f):
        @functools.wraps(f)
        async def call(self, *args, **kwargs):
            r = await self.grpc_call(method(self), request(*args, **kwargs))
            return response_func(r)

        return call

    return _f


class Auth(StubMixin):
    def _update_channel(self, channel):
        super()._update_channel(channel)
        self._auth_stub = stub.AuthStub(channel)

    @call_grpc(lambda: rpc.AuthEnableRequest(), lambda r: None, lambda s: s._auth_stub.AuthEnable)
    async def auth_enable(self):
        pass

    @call_grpc(lambda: rpc.AuthDisableRequest(), lambda r: None, lambda s: s._auth_stub.AuthDisable)
    async def auth_disable(self):
        pass

    @call_grpc(lambda username, password: rpc.AuthenticateRequest(name=username, password=password),
               lambda r: r.token, lambda s: s._auth_stub.Authenticate)
    async def authenticate(self, username, password):
        pass

    @call_grpc(lambda: rpc.AuthUserListRequest(), lambda r: [u for u in r.users], lambda s: s._auth_stub.UserList)
    async def user_list(self):
        pass

    @call_grpc(lambda username: rpc.AuthUserGetRequest(name=username), lambda r: [r for r in r.roles],
               lambda s: s._auth_stub.UserGet)
    async def user_get(self, username):
        pass

    @call_grpc(lambda username, password: rpc.AuthUserAddRequest(name=username, password=password), lambda r: None,
               lambda s: s._auth_stub.UserAdd)
    async def user_add(self, username, password):
        pass

    @call_grpc(lambda username: rpc.AuthUserDeleteRequest(name=username), lambda r: None,
               lambda s: s._auth_stub.UserDelete)
    async def user_delete(self, username):
        pass

    @call_grpc(lambda username, password: rpc.AuthUserChangePasswordRequest(name=username, password=password),
               lambda r: None, lambda s: s._auth_stub.UserChangePassword)
    async def user_change_password(self, username, password):
        pass

    @call_grpc(lambda username, role: rpc.AuthUserGrantRoleRequest(user=username, role=role), lambda r: None,
               lambda s: s._auth_stub.UserGrantRole)
    async def user_grant_role(self, username, role):
        pass

    @call_grpc(lambda username, role: rpc.AuthUserRevokeRoleRequest(name=username, role=role), lambda r: None,
               lambda s: s._auth_stub.UserRevokeRole)
    async def user_revoke_role(self, username, role):
        pass

    @call_grpc(lambda: rpc.AuthRoleListRequest(), lambda r: [role for role in r.roles],
               lambda s: s._auth_stub.RoleList)
    async def role_list(self):
        pass

    @call_grpc(lambda name: rpc.AuthRoleGetRequest(role=name), lambda r: [p for p in r.perm],
               lambda s: s._auth_stub.RoleGet)
    async def role_get(self, name):
        pass

    @call_grpc(lambda name: rpc.AuthRoleAddRequest(name=name), lambda r: None, lambda s: s._auth_stub.RoleAdd)
    async def role_add(self, name):
        pass

    @call_grpc(lambda name: rpc.AuthRoleDeleteRequest(role=name), lambda r: None, lambda s: s._auth_stub.RoleDelete)
    async def role_delete(self, name):
        pass

    @staticmethod
    def role_grant_request(name, key_range, permission):
        if permission not in [auth.Permission.READ, auth.Permission.WRITE, auth.Permission.READWRITE]:
            raise ValueError("permission must be read, write or readwrite")
        per = auth.Permission(permType=permission)
        put_key_range(per, key_range)

        request = rpc.AuthRoleGrantPermissionRequest(name=name, perm=per)

        return request

    @call_grpc(role_grant_request.__func__, lambda r: None, lambda s: s._auth_stub.RoleGrantPermission)
    async def role_grant_permission(self, name, key_range, permission):
        pass

    @staticmethod
    def role_revoke_request(name, key_range):
        request = rpc.AuthRoleRevokePermissionRequest(role=name)
        put_key_range(request, key_range)

        return request

    @call_grpc(role_revoke_request.__func__, lambda r: None, lambda s: s._auth_stub.RoleRevokePermission)
    async def role_revoke_permission(self, name, key_range):
        pass



import unittest
import asyncio
import functools

from aioetcd3.client import client
from aioetcd3.help import range_all


def asynctest(f):
    @functools.wraps(f)
    def _f(self):
        asyncio.get_event_loop().run_until_complete(f(self))

    return _f


class ClusterTest(unittest.TestCase):
    def setUp(self):
        endpoints = "127.0.0.1:2379"
        self.client = client(endpoint=endpoints)

    @asynctest
    async def test_member(self):
        members = await self.client.member_list()
        self.assertTrue(members)

        m = members[0]
        # urls = [u for u in m.clientURLs]
        # urls = [u.rpartition("//")[2] for u in urls]

        healthy, unhealthy = await self.client.member_healthy([m.clientURLs])
        self.assertTrue(healthy)
        self.assertFalse(unhealthy)

        healthy, unhealthy = await self.client.member_healthy()
        self.assertTrue(healthy)
        self.assertFalse(unhealthy)




from aioetcd3._etcdv3 import rpc_pb2 as rpc
from aioetcd3.base import StubMixin

import functools
import inspect
import asyncio
import aioetcd3._etcdv3.rpc_pb2_grpc as stub


def call_grpc(request, response_func, method):

    def _f(f):
        @functools.wraps(f)
        async def call(self, *args, **kwargs):
            params = inspect.getcallargs(f, self, *args, **kwargs)
            params.pop('self')
            r = await self.grpc_call(method(self), request(**params))
            return response_func(r, client=self)

        return call

    return _f


class RLease(object):
    def __init__(self, ttl, id, client):
        self.ttl = ttl
        self.id = id
        self.client = client

    async def __aenter__(self):
        lease = await self.client.grant_lease(ttl=self.ttl)
        self.ttl = lease.ttl
        self.id = lease.id

        refresh_ttl = self.ttl // 2

        async def task(cycle):
            while True:
                await asyncio.sleep(cycle)
                await self.refresh()

        self.refresh_task = asyncio.ensure_future(task(refresh_ttl))
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if hasattr(self, 'refresh_task'):
            self.refresh_task.cancel()
            asyncio.wait(self.refresh_task)
        await self.revoke()

    async def revoke(self):
        return await self.client.revoke_lease(self.id)

    async def refresh(self):
        return await self.client.refresh_lease(self.id)

    async def info(self):
        return await self.client.get_lease_info(self.id)


class Lease(StubMixin):

    def _update_channel(self, channel):
        super()._update_channel(channel)
        self._lease_stub = stub.LeaseStub(channel)

    @call_grpc(lambda ttl, id: rpc.LeaseGrantRequest(TTL=ttl, ID=id),
               lambda r, client: RLease(r.TTL, r.ID, client),
               lambda s: s._lease_stub.LeaseGrant)
    async def grant_lease(self, ttl, id=0):
        pass

    def grant_lease_scope(self, ttl, id=0):
        return RLease(ttl, id, self)

    @call_grpc(lambda lease: rpc.LeaseRevokeRequest(ID=get_lease_id(lease)),
               lambda r, client: None, lambda s: s._lease_stub.LeaseRevoke)
    async def revoke_lease(self, lease):
        pass

    async def refresh_lease(self, lease):
        lease_id = get_lease_id(lease)
        lease_request = rpc.LeaseKeepAliveRequest(ID=lease_id)

        async def generate_request(request):
            for re in [request]:
                yield re

        new_lease = None
        async with self._lease_stub.LeaseKeepAlive.with_scope(generate_request(lease_request)) as result:
            async for r in result:
                self._update_cluster_info(r.header)
                new_lease = RLease(r.TTL, r.ID, self)

        return new_lease

    @call_grpc(lambda lease: rpc.LeaseTimeToLiveRequest(ID=get_lease_id(lease), keys=True),
               lambda r, client: (RLease(r.TTL, r.ID, client), [k for k in r.keys]) if r.TTL >= 0 else (None, []),
               lambda s: s._lease_stub.LeaseTimeToLive)
    async def get_lease_info(self, lease):
        pass


def get_lease_id(lease):
    if hasattr(lease, 'id'):
        return lease.id
    else:
        return lease




from aioetcd3._etcdv3 import rpc_pb2 as rpc
from aioetcd3.utils import to_bytes, put_key_range
from aioetcd3.base import StubMixin, _default_timeout
from inspect import getcallargs
import functools
import aioetcd3._etcdv3.rpc_pb2_grpc as stub


class KVMetadata(object):
    def __init__(self, keyvalue):
        # self.key = keyvalue.key
        self.create_revision = keyvalue.create_revision
        self.mod_revision = keyvalue.mod_revision
        self.version = keyvalue.version
        self.lease_id = keyvalue.lease

_default = object()


_sort_order_dict = {"ascend": rpc.RangeRequest.ASCEND,
                    None: rpc.RangeRequest.NONE,
                    "descend": rpc.RangeRequest.DESCEND}

_sort_target_dict = {"key": rpc.RangeRequest.KEY,
                     None: rpc.RangeRequest.KEY,
                     'version': rpc.RangeRequest.VERSION,
                     'create': rpc.RangeRequest.CREATE,
                     'mod': rpc.RangeRequest.MOD,
                     'value': rpc.RangeRequest.VALUE}


def _get_grpc_args(func, *args, **kwargs):
    params = getcallargs(func, None, *args, **kwargs)
    params.pop('self')
    params.pop('timeout')
    return params


def _kv(request_builder, response_builder, method):
    def _decorator(f):
        def txn(*args, timeout=_default_timeout, **kwargs):
            call_args = _get_grpc_args(f, *args, **kwargs)
            return (request_builder(**call_args), response_builder(**call_args))
        f.txn = txn

        @functools.wraps(f)
        async def grpc_func(self, *args, timeout=_default_timeout, **kwargs):
            request, response = txn(*args, **kwargs)
            return response(await self.grpc_call(method(self), request, timeout=timeout))
        return grpc_func
    return _decorator


def _create_txn_response_builder(success, fail, **kwargs):
    def _response_builder(response):
        if response.succeeded:
            return True, [t[1](_get_op_response(r)) for t, r in zip(success, response.responses)]
        else:
            return False, [t[1](_get_op_response(r)) for t, r in zip(fail, response.responses)]
    return _response_builder



def _range_request(key_range, sort_order=None, sort_target='key', **kwargs):
    range_request = rpc.RangeRequest()
    put_key_range(range_request, key_range)

    for k, v in kwargs.items():
        if v is not None:
            setattr(range_request, k, v)

    if sort_order in _sort_order_dict:
        range_request.sort_order = _sort_order_dict[sort_order]
    else:
        raise ValueError('unknown sort order: "{}"'.format(sort_order))

    if sort_target in _sort_target_dict:
        range_request.sort_target=_sort_target_dict[sort_target]
    else:
        raise ValueError('sort_target must be one of "key", '
                         '"version", "create", "mod" or "value"')

    return range_request


def _range_response(kv_response):
    result = []
    for kv in kv_response.kvs:
        result.append((kv.key, kv.value, KVMetadata(kv)))
    return result


def _static_builder(f):
    def _builder(*args, **kwargs):
        return f
    return _builder


def _partial_builder(f):
    def _builder(**kwargs):
        return functools.partial(f, **kwargs)
    return _builder


def _put_request(key, value, lease=None, prev_kv=False, ignore_value=False, ignore_lease=False):
    if lease is None:
        lease = 0
    elif hasattr(lease, 'id'):
        lease = lease.id
    put_request = rpc.PutRequest(key=to_bytes(key),
                                 prev_kv=prev_kv, ignore_value=ignore_value,
                                 ignore_lease=ignore_lease)

    if not ignore_value:
        put_request.value = to_bytes(value)

    if not ignore_lease:
        put_request.lease = lease

    return put_request


def _delete_request(key_range, prev_kv=False):

    delete_request = rpc.DeleteRangeRequest(prev_kv=prev_kv)
    put_key_range(delete_request, key_range)

    return delete_request


def _get_response(response):
    if response.kvs:
        return response.kvs[0].value, KVMetadata(response.kvs[0])
    else:
        return None, None


def _range_keys_response(response):
    result = []
    for kv in response.kvs:
        result.append((kv.key, KVMetadata(kv)))

    return result


def _delete_response(response, prev_kv=False, **kwargs):
    # when set prev_kv to return prev value,
    # but it is not existed , response has no prev_kvs
    if prev_kv:
        r = []
        for kv in response.prev_kvs:
            r.append((kv.key, kv.value, KVMetadata(kv)))
        return r
    else:
        return response.deleted


def _put_response(response, prev_kv=False, **kwargs):

    # when set prev_kv to return prev value,
    # but it is not existed , response has no prev_kv
    if prev_kv and response.HasField('prev_kv'):
        return response.prev_kv.value, KVMetadata(response.prev_kv)
    else:
        return None, None


def _create_op_request(request):
    if isinstance(request, rpc.PutRequest):
        return rpc.RequestOp(request_put=request)
    elif isinstance(request, rpc.RangeRequest):
        return rpc.RequestOp(request_range=request)
    elif isinstance(request, rpc.DeleteRangeRequest):
        return rpc.RequestOp(request_delete_range=request)
    elif isinstance(request, rpc.TxnRequest):
        return rpc.RequestOp(request_txn=request)
    else:
        raise TypeError("Unsupported request OP: " + repr(request))


def _get_op_response(response):
    return getattr(response, response.WhichOneof('response'))


def _compare_request(compare, success, fail):
    compare_message = [c.build_message() for c in compare]
    success_message = [_create_op_request(request=r) for r, _ in success]
    fail_message = [_create_op_request(request=r) for r, _ in fail]
    request = rpc.TxnRequest(compare=compare_message, success=success_message, failure=fail_message)
    return request


class KV(StubMixin):
    def _update_channel(self, channel):
        super()._update_channel(channel)
        self._kv_stub = stub.KVStub(channel)

    @_kv(_range_request, _static_builder(_range_response), lambda x: x._kv_stub.Range)
    async def range(self, key_range, limit=None, revision=None, timeout=_default_timeout, sort_order=None, sort_target='key',
                    serializable=None, keys_only=None, count_only=None, min_mod_revision=None, max_mod_revision=None,
                    min_create_revision=None, max_create_revision=None):
        # implemented in decorator
        pass

    @_kv(functools.partial(_range_request, count_only=True),
         _static_builder(lambda r: r.count), lambda x: x._kv_stub.Range)
    async def count(self, key_range, revision=None, timeout=_default_timeout, min_mod_revision=None,
                    max_mod_revision=None, min_create_revision=None, max_create_revision=None):
        pass

    @_kv(functools.partial(_range_request, keys_only=True), _static_builder(_range_keys_response),
         lambda x: x._kv_stub.Range)
    async def range_keys(self, key_range, limit=None, revison=None, sort_order=None,
                         sort_target='key', timeout=_default_timeout, serializable=None, count_only=None,
                         min_mod_revision=None, max_mod_revision=None, min_create_revision=None,
                         max_create_revision=None):
        pass

    @_kv(_range_request, _static_builder(_get_response), lambda x: x._kv_stub.Range)
    async def get(self, key_range, revision=None, timeout=_default_timeout, serializable=None,
                  min_mod_revision=None, max_mod_revision=None, min_create_revision=None,
                  max_create_revision=None):
        pass

    @_kv(_put_request, _partial_builder(_put_response), lambda x: x._kv_stub.Put)
    async def put(self, key, value, lease=0, prev_kv=False, timeout=_default_timeout, ignore_value=False, ignore_lease=False):
        pass

    @_kv(_delete_request, _partial_builder(_delete_response), lambda x: x._kv_stub.DeleteRange)
    async def delete(self, key_range, timeout=_default_timeout, prev_kv=False):
        pass

    @_kv(functools.partial(_delete_request, prev_kv=True),
         _partial_builder(functools.partial(_delete_response, prev_kv=True)),
         lambda x: x._kv_stub.DeleteRange)
    async def pop(self, key_range, timeout=_default_timeout):
        pass

    @_kv(_compare_request, _create_txn_response_builder, lambda x: x._kv_stub.Txn)
    async def txn(self, compare, success, fail=[], *, timeout=_default_timeout):
        pass

    async def compact(self, revision, physical=False, *, timeout=_default_timeout):
        """
        Compact etcd KV storage
        
        :param revision: compact to specified revision
        
        :param physical: return until data is physically compacted
        
        :param timeout: maximum time to wait
        """
        await self.grpc_call(self._kv_stub.Compact,
                               rpc.CompactionRequest(revision=revision,
                                                    physical=physical),
                               timeout=timeout)
        


import functools

from aioetcd3._etcdv3 import rpc_pb2 as rpc
from aioetcd3.base import StubMixin
import aioetcd3._etcdv3.rpc_pb2_grpc as stub


class Status(object):
    def __init__(self, status):
        self.version = status.version
        self.dbSize = status.dbSize
        self.leader = status.leader
        self.raftIndex = status.raftIndex
        self.raftTerm = status.raftTerm


def call_grpc(request, response_func, method):

    def _f(f):
        @functools.wraps(f)
        async def call(self, *args, **kwargs):
            r = await self.grpc_call(method(self), request(*args, **kwargs))
            return response_func(r)

        return call

    return _f


class Maintenance(StubMixin):
    def _update_channel(self, channel):
        super()._update_channel(channel)
        self._maintenance_stub = stub.MaintenanceStub(channel)

    @call_grpc(lambda: rpc.StatusRequest(), lambda r: Status(r),
               lambda s: s._maintenance_stub.Status)
    async def status(self):
        pass

    @call_grpc(lambda action, type, mid: rpc.AlarmRequest(action=action, memberID=mid, alarm=type),
               lambda r: [m for m in r.alarms], lambda s: s._maintenance_stub.Alarm)
    async def alarm(self, action, type, mid=0):
        pass

    @call_grpc(lambda: rpc.SnapshotRequest(), lambda r: (r.remaining_bytes, r.blob),
               lambda s: s._maintenance_stub.Snapshot)
    async def snapshot(self):
        pass

    @call_grpc(lambda: rpc.HashRequest(), lambda r: r.hash,
               lambda s: s._maintenance_stub.Hash)
    async def hash(self):
        pass

    @call_grpc(lambda: rpc.DefragmentRequest(), lambda r: None,
               lambda s: s._maintenance_stub.Defragment)
    async def defragment(self):
        pass



import aiogrpc
import os
from aiogrpc.channel import Channel
from aioetcd3.kv import KV
from aioetcd3.lease import Lease
from aioetcd3.auth import Auth
from aioetcd3.watch import Watch
from aioetcd3.maintenance import Maintenance
from aioetcd3.cluster import Cluster
from aioetcd3.utils import get_secure_creds


class Client(KV, Lease, Auth, Watch, Maintenance, Cluster):
    def __init__(self, endpoint, ssl=False,
                 ca_cert=None, cert_key=None, cert_cert=None,
                 default_ca=False, grpc_options = None, timeout=5,
                 *, loop=None, executor=None):
        channel = self._create_grpc_channel(endpoint=endpoint, ssl=ssl,
                                            ca_cert=ca_cert,
                                            cert_key=cert_key, cert_cert=cert_cert,
                                            default_ca=default_ca,
                                            options=grpc_options,
                                            loop=loop,
                                            executor=executor)
        super().__init__(channel, timeout)

    def update_server_list(self, endpoint):
        self.close()
        channel = self._recreate_grpc_channel(endpoint)
        self._update_channel(channel)

    def _create_grpc_channel(self, endpoint, ssl=False,
                             ca_cert=None, cert_key=None, cert_cert=None, default_ca=False, options=None,
                             *, loop=None, executor=None):
        credentials = None
        if not ssl:
            channel = aiogrpc.insecure_channel(endpoint, options=options, loop=loop, executor=executor,
                                               standalone_pool_for_streaming=True)
        else:
            if default_ca:
                ca_cert = None
            else:
                if ca_cert is None:
                    ca_cert = ''

            # to ensure ssl connect , set grpc env
            # os.environ['GRPC_SSL_CIPHER_SUITES'] = 'ECDHE-ECDSA-AES256-GCM-SHA384'

            credentials = aiogrpc.ssl_channel_credentials(ca_cert, cert_key, cert_cert)
            channel = aiogrpc.secure_channel(endpoint, credentials, options=options,
                                             loop=loop, executor=executor,
                                             standalone_pool_for_streaming=True)

        # Save parameters for auto-recreate
        self._credentials = credentials
        self._options = options
        self._loop = channel._loop
        self._executor = executor
        return channel

    def _recreate_grpc_channel(self, endpoint):
        if self._credentials:
            channel = aiogrpc.secure_channel(endpoint, self._credentials, options=self._options,
                                             loop=self._loop, executor=self._executor,
                                             standalone_pool_for_streaming=True)
        else:
            channel = aiogrpc.insecure_channel(endpoint, options=self._options, loop=self._loop,
                                               executor=self._executor, standalone_pool_for_streaming=True)
        return channel
    
    def close(self):
        return self.channel.close()


def client(endpoint, grpc_options=None, timeout=None):

    # user `ip:port,ip:port` to user grpc balance
    return Client(endpoint, grpc_options=grpc_options, timeout=timeout)


def ssl_client(endpoint, ca_file=None, cert_file=None, key_file=None, default_ca=False, grpc_options=None,
               timeout=None):
    ca, key, cert = get_secure_creds(ca_cert=ca_file, cert_cert=cert_file, cert_key=key_file)
    return Client(endpoint, ssl=True, ca_cert=ca, cert_key=key, cert_cert=cert,
                  default_ca=default_ca, grpc_options=grpc_options, timeout=timeout)


def set_grpc_cipher(enable_rsa=True, enable_ecdsa=True, ciphers=None):
    """
    Set GRPC_SSL_CIPHER_SUITES environment variable to change the SSL cipher
    used by GRPC. By default the GRPC C core only supports RSA.

    :param enable_rsa:  Enable RSA cipher
    :param enable_ecdsa: Enable ECDSA cipher
    :param ciphers: Override the cipher list to a list of strings
    """
    if ciphers:
        os.environ['GRPC_SSL_CIPHER_SUITES'] = ':'.join(ciphers)
    else:
        rsa_ciphers = 'ECDHE-RSA-AES128-GCM-SHA256:ECDHE-RSA-AES128-SHA256:' \
                      'ECDHE-RSA-AES256-SHA384:ECDHE-RSA-AES256-GCM-SHA384'
        ecdsa_ciphers = 'ECDHE-ECDSA-AES256-GCM-SHA384'
        if enable_rsa:
            if enable_ecdsa:
                env = rsa_ciphers + ':' + ecdsa_ciphers
            else:
                env = rsa_ciphers
        else:
            if enable_ecdsa:
                env = ecdsa_ciphers
            else:
                env = None
        if env is None:
            if 'GRPC_SSL_CIPHER_SUITES' in os.environ:
                del os.environ['GRPC_SSL_CIPHER_SUITES']
        else:
            os.environ['GRPC_SSL_CIPHER_SUITES'] = env



import unittest
import asyncio
import functools

from aioetcd3.client import client, ssl_client, set_grpc_cipher
from aioetcd3.help import range_all, PER_RW


def asynctest(f):
    @functools.wraps(f)
    def _f(self):
        asyncio.get_event_loop().run_until_complete(f(self))

    return _f

TEST_USER_NAME = 'test'
TEST_ROLE_NAME = 'admin'


class AuthTest(unittest.TestCase):
    def setUp(self):
        endpoints = "127.0.0.1:2379"
        self.client = client(endpoint=endpoints)

        set_grpc_cipher()
        auth_etcd_url = "127.0.0.1:2378"
        self.root_client = ssl_client(endpoint=auth_etcd_url, ca_file="test/cfssl/ca.pem",
                                      cert_file="test/cfssl/client-root.pem",
                                      key_file="test/cfssl/client-root-key.pem")

        self.client_client = ssl_client(endpoint=auth_etcd_url, ca_file="test/cfssl/ca.pem",
                                        cert_file="test/cfssl/client.pem",
                                        key_file="test/cfssl/client-key.pem")

        self.tearDown()

    @asynctest
    async def test_auth_1(self):

        await self.client.user_add(username=TEST_USER_NAME, password='1234')
        users = await self.client.user_list()

        self.assertIn(TEST_USER_NAME, users)

        roles = await self.client.user_get(username=TEST_USER_NAME)
        self.assertEqual(len(roles), 0)

        await self.client.user_change_password(username=TEST_USER_NAME, password="test")

        await self.client.user_delete(username=TEST_USER_NAME)

    @asynctest
    async def test_auth_2(self):

        await self.client.role_add(name=TEST_ROLE_NAME)

        roles = await self.client.role_list()
        self.assertIn(TEST_ROLE_NAME, roles)

        role_info = await self.client.role_get(name=TEST_ROLE_NAME)

        await self.client.role_delete(name=TEST_ROLE_NAME)

    @asynctest
    async def test_auth_3(self):

        await self.client.user_add(username=TEST_USER_NAME, password="test")
        with self.assertRaises(Exception):
            await self.client.user_grant_role(username=TEST_USER_NAME, role=TEST_ROLE_NAME)

        await self.client.role_add(name=TEST_ROLE_NAME)
        await self.client.user_grant_role(username=TEST_USER_NAME, role=TEST_ROLE_NAME)

        await self.client.role_grant_permission(name=TEST_ROLE_NAME,
                                                key_range=range_all(),
                                                permission=PER_RW)

        await self.client.user_revoke_role(username=TEST_USER_NAME, role=TEST_ROLE_NAME)

        await self.client.role_revoke_permission(name=TEST_ROLE_NAME,
                                                 key_range=range_all())


    @asynctest
    async def test_auth_4(self):
        await self.root_client.user_add(username='root', password='root')
        await self.root_client.role_add(name='root')
        await self.root_client.user_grant_role(username='root', role='root')
        await self.root_client.auth_enable()

        await self.root_client.user_add(username='client', password='client')
        await self.root_client.role_add(name='client')

        await self.root_client.put('/foo', '/foo')
        value, meta = await self.root_client.get('/foo')
        self.assertEqual(value, b'/foo')

        with self.assertRaises(Exception):
            await self.client_client.get('/foo')

        await self.root_client.role_grant_permission(name='client', key_range='/foo', permission=PER_RW)
        await self.root_client.user_grant_role(username='client', role='client')

        value, meta = await self.client_client.get('/foo')
        self.assertEqual(value, b'/foo')

        await self.client_client.put('/foo', 'ssss')

    async def delete_all_user(self):
        users = await self.client.user_list()

        for u in users:
            await self.client.user_delete(username=u)

        users = await self.root_client.user_list()

        for u in users:
            await self.root_client.user_delete(username=u)

    async def delete_all_role(self):
        roles = await self.client.role_list()

        for r in roles:
            await self.client.role_delete(name=r)

        roles = await self.root_client.role_list()

        for r in roles:
            await self.root_client.role_delete(name=r)

    @asynctest
    async def tearDown(self):

        await self.client.delete(range_all())

        await self.root_client.auth_disable()

        await self.delete_all_user()
        await self.delete_all_role()




from aioetcd3.utils import increment_last_byte, to_bytes, next_valid_key
from aioetcd3._etcdv3 import auth_pb2 as _auth
from aioetcd3._etcdv3 import rpc_pb2 as _rpc

SORT_ASCEND = 'ascend'
SORT_DESCEND = 'descend'

PER_R = _auth.Permission.READ
PER_W = _auth.Permission.WRITE
PER_RW = _auth.Permission.READWRITE

ALARM_ACTION_GET = _rpc.AlarmRequest.GET
ALARM_ACTION_ACTIVATE = _rpc.AlarmRequest.ACTIVATE
ALARM_ACTION_DEACTIVATE = _rpc.AlarmRequest.DEACTIVATE

ALARM_TYPE_NONE = _rpc.NONE
ALARM_TYPE_NOSPACE = _rpc.NOSPACE


def range_prefix(key):
    if not key:
        return range_all()
    else:
        return to_bytes(key), increment_last_byte(to_bytes(key))


def range_prefix_excluding(prefix, with_out):
    """
    Return a list of key_range, union of which is a prefix range excluding some keys

    :param prefix: the key to generate the range prefix
    :param with_out: a list of key_range (key or (start,end) tuple)
    :return: a list of key_range, union of which is a prefix range excluding some keys
    """
    return range_excluding(range_prefix(prefix), with_out)


def range_excluding(range_, with_out):
    """
    Return a list of key_range, union of which is a range excluding some keys

    :param range_: the original range
    :param with_out: a list of key_range (key or (start,end) tuple)
    :return: a list of key_range, union of which is a prefix range excluding some keys
    """
    # Merge with_out
    with_out_ranges = [(to_bytes(v), next_valid_key(v)) if isinstance(v, str) or isinstance(v, bytes)
                       else (to_bytes(v[0]), to_bytes(v[1]))
                       for v in with_out]
    with_out_ranges.sort()
    range_start, range_end = range_
    range_start = to_bytes(range_start)
    range_end = to_bytes(range_end)
    re_range = []
    next_start_key = range_start
    for s, e in with_out_ranges:
        if s >= range_end != b'\x00':
            break
        start, end = next_start_key, s
        if start < end:
            re_range.append((start, end))
        if e == b'\x00':
            next_start_key = None
            break
        else:
            next_start_key = max(next_start_key, e)
    if next_start_key is not None and \
            (next_start_key < range_end or
                range_end == b'\x00'):
        re_range.append((next_start_key, range_end))
    return re_range


def range_greater(key):
    return next_valid_key(key), b'\0'


def range_greater_equal(key):
    return key, b'\0'


def range_less(key):
    return b'\0', key


def range_less_equal(key):
    return b'\0', next_valid_key(key)


def range_all():
    return b'\0', b'\0'



def to_bytes(maybe_bytestring):
    """
    Encode string to bytes.

    Convenience function to do a simple encode('utf-8') if the input is not
    already bytes. Returns the data unmodified if the input is bytes.
    """
    if isinstance(maybe_bytestring, bytes):
        return maybe_bytestring
    else:
        return maybe_bytestring.encode('utf-8')


def increment_last_byte(byte_string):
    s = bytearray(to_bytes(byte_string))
    for i in range(len(s) - 1, -1, -1):
        if s[i] < 0xff:
            s[i] += 1
            return bytes(s[:i+1])
    else:
        return b'\x00'


def next_valid_key(byte_string):
    return to_bytes(byte_string) + b'\x00'


def put_key_range(obj, key_range):
    if isinstance(key_range, str) or isinstance(key_range, bytes):
        obj.key = to_bytes(key_range)
    else:
        try:
            key, range_end = key_range
        except Exception:
            raise ValueError("key_range must be either a str/bytes 'key', or ('key', 'range_end') tuple")
        obj.key = to_bytes(key)
        obj.range_end = to_bytes(range_end)
    return obj


def ipv4_endpoints(server_list):
    return 'ipv4:///' + ','.join(
                    s + ':2379' if ':' not in s else s
                    for s in server_list
            )


def ipv6_endpoints(server_list):

    format_server_list = []
    for ip_address in server_list:
        if ip_address.startswith('['):
            if ip_address.endswith(']'):
                ip_address = ip_address + ':2379'
        else:
            if not ip_address.endswith(']'):
                ip_address = '[' + ip_address + ']:2379'

        format_server_list.append(ip_address)

    return 'ipv6:///' + ','.join(format_server_list)


def dns_endpoint(dns_name):
    return 'dns:///' + dns_name


def get_secure_creds(ca_cert, cert_key, cert_cert):
    with open(ca_cert, 'rb') as ca_cert_file:
        with open(cert_key, 'rb') as cert_key_file:
            with open(cert_cert, 'rb') as cert_cert_file:
                return ca_cert_file.read(), cert_key_file.read(), cert_cert_file.read()

